/*
 *  LibRoadnav
 *  wxAntiAliasedDC.h
 *
 *  Copyright (c) 2004 - 2007 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// wxAntiAliasedDC wraps around an existing wxDC class and makes various
/// function calls produce antialiased output.
///
//////////////////////////////////////////////////////////////////////////////

#ifndef _WXANTIALIASEDDC_H
#define _WXANTIALIASEDDC_H

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include <wx/image.h>

#include <map>

//////////////////////////////////////////////////////////////////////////////
///
/// \brief wxAntiAliasedDC wraps around an existing wxDC class and makes various
/// function calls produce antialiased output.
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
class wxAntiAliasedDC : public T
{
	public:

#ifdef WXAADC_CONSTRUCTOR_ARGS
		wxAntiAliasedDC(wxWindow * pWnd);
#else
		wxAntiAliasedDC();
#endif
		~wxAntiAliasedDC();

		virtual void SetBackground(const wxBrush& brush);
		
#if wxCHECK_VERSION(2, 8, 0)
		void DoSelect(const wxBitmap & bmpObj);
#else
		void SelectObject(const wxBitmap & bmpObj);
#endif
		
		void Setup(int iWidth, int iHeight);
	
	protected:
	    virtual void DoDrawLine(wxCoord x1, wxCoord y1, wxCoord x2, wxCoord y2);
	    virtual void DoDrawPolygon(int n, wxPoint points[],
                               wxCoord xoffset, wxCoord yoffset,
                               int fillStyle = wxODDEVEN_RULE);
		virtual void DoDrawText(const wxString& text, wxCoord x, wxCoord y);
		virtual void DoDrawRotatedText(const wxString& text,
									   wxCoord x, wxCoord y, double angle);
		virtual void DoDrawEllipse(wxCoord x, wxCoord y,
								   wxCoord width, wxCoord height);
		virtual void DoGetTextExtent(const wxString& string,
									 wxCoord *x, wxCoord *y,
									 wxCoord *descent = NULL,
									 wxCoord *externalLeading = NULL,
									 wxFont *theFont = NULL) const;
	    virtual void DoDrawPoint(wxCoord x, wxCoord y);
		virtual void DoDrawLines(int n, wxPoint points[], wxCoord xoffset, wxCoord yoffset);
		
	private:
		void ShadePoint(wxCoord x, wxCoord y, const wxColour & c, double fWeight);
		
		wxColour GetPixelColor(wxCoord x, wxCoord y);
		void SetPixelColor(wxCoord x, wxCoord y, const wxColour & c);

		int m_iHeight;
		int m_iWidth;
		wxUint8 * m_pcPixelCache;
	
		int m_iAALevel;
	
		wxColour m_clrMask;
		wxPen m_penMask;
		wxBrush m_brMask;
};

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - initialize the masking pen and brush
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
#ifdef WXAADC_CONSTRUCTOR_ARGS
wxAntiAliasedDC<T>::wxAntiAliasedDC(wxWindow * pWnd) : T(pWnd)
#else
wxAntiAliasedDC<T>::wxAntiAliasedDC() : T()
#endif
{
	m_iAALevel = 3;
	
	m_iHeight = 0;
	m_iWidth = 0;
	m_pcPixelCache = NULL;
	
	m_clrMask.Set(254, 254, 254);
	
	m_penMask = wxPen(m_clrMask, 0, wxTRANSPARENT);
	m_brMask = wxBrush(m_clrMask, wxSOLID);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
wxAntiAliasedDC<T>::~wxAntiAliasedDC()
{
	if (m_pcPixelCache)
		delete [] m_pcPixelCache;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Overrides default SetBackground call - updates masking pen and brush,
/// then calls base class
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::SetBackground(const wxBrush& brush)
{
	T::SetBackground(brush);

	// catch any truncation due to 16bpp color depth
	m_clrMask = T::GetBackground().GetColour();

	// the mask color should not be used anywhere, otherwise it will appear clear
	m_clrMask.Set(
						m_clrMask.Red() > 247 ? 240 : m_clrMask.Red() + 8,
						m_clrMask.Green() > 247 ? 240 : m_clrMask.Green() + 8,
						m_clrMask.Blue() > 247 ? 240 : m_clrMask.Blue() + 8
				);
	
	m_penMask = wxPen(m_clrMask, 0, wxTRANSPARENT);
	m_brMask = wxBrush(m_clrMask, wxSOLID);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Overrides default DoDrawPoint - remembers the stored color then calls the
/// base class
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::DoDrawPoint(wxCoord x, wxCoord y)
{
	if (x < 0 || y < 0 || x >= m_iWidth || y >= m_iHeight)
		return;

	SetPixelColor(x, y, T::GetPen().GetColour());
	
	T::DoDrawPoint(x, y);
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Tint a pixel a certain color
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::ShadePoint(wxCoord x, wxCoord y, const wxColour & c, double fWeight)
{
	if (x < 0 || y < 0 || x >= m_iWidth || y >= m_iHeight)
		return;

	// GetPixel is really slow! Using a cache instead
	// not quite as good, but it will do for Roadnav since lines
	// are drawn before everything else	
	//GetPixel(x, y, &clrOld);

	wxColour clrOld;

	clrOld = GetPixelColor(x, y);
	
	wxPen penNew(		wxColour(	(int) (clrOld.Red() * (1 - fWeight) + c.Red() * fWeight),
									(int) (clrOld.Green() * (1 - fWeight) + c.Green() * fWeight),
									(int) (clrOld.Blue() * (1 - fWeight) + c.Blue() * fWeight)
								),
						1, 
						wxSOLID);
	
	T::SetPen(penNew);
	
	T::DrawPoint(x, y);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Replaces default DoDrawLine - draws an antialiased line
///
/// Limitations:
/// - Among other things, only supports solid and dotted lines
/// - Line width isn't exactly right
/// - End of line isn't drawn quite right (it's drawn with a horizontal or
///   vertical line edge rather than an edge perpendicular to the line)
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::DoDrawLine(wxCoord x1, wxCoord y1, wxCoord x2, wxCoord y2)
{
	int iDX, iDY;

	wxPen penCurrent;
	wxColour clrPen;
	int iPenWidth;
	int iPenStyle;
	
	iDX = x2 - x1;
	iDY = y2 - y1;
	
	if (iDX == 0 && iDY == 0)
	{
		T::DrawPoint(x1, y1);
		return;
	}

	// Fixes stack overflow under OS X (DoDrawPoint calls DoDrawLine for some odd reason)
	if ((abs(iDX) + abs(iDY)) <= 4)
	{
		T::DoDrawLine(x1, y1, x2, y2);
		return;
	}
	
	penCurrent = T::GetPen();
	clrPen = penCurrent.GetColour();
	iPenWidth = penCurrent.GetWidth();
	iPenStyle = penCurrent.GetStyle();

	if (abs(iDX) > abs(iDY))
	{
		int iStep;
		int iX;
		bool bDone;
		
		iStep = 1;
		if (x1 > x2)
			iStep = -1;
		
		if (iPenStyle != wxSOLID)
			iStep *= 3;
		
		bDone = false;
		
		for (iX = x1; !bDone; iX += iStep)
		{
			double fY;
			double fWeightTop;
			double fWeightBottom;
			int iShift;
			int iTop;
			int iBottom;
			
			fY = ((double) (iX - x1) / iDX * iDY) + y1;
			
			fWeightBottom = fY - floor(fY);
			fWeightTop = 1 - fWeightBottom;

			wxASSERT(fWeightBottom <= 1);
			wxASSERT(fWeightBottom >= 0);

			wxASSERT(fWeightTop <= 1);
			wxASSERT(fWeightTop >= 0);
			
			iTop = -iPenWidth / 2;
			iBottom = iTop + iPenWidth - 1;
			
			iShift = iTop;
			ShadePoint(iX, (int) floor(fY + iShift), clrPen, fWeightTop);
			
			T::SetPen(penCurrent);
			for (iShift++; iShift <= iBottom; iShift++)
				T::DrawPoint(iX, (int) floor(fY + iShift));

			ShadePoint(iX, (int) floor(fY + iShift), clrPen, fWeightBottom);
			
			if (x2 > x1 && iX >= x2)
				bDone = true;
			if (x2 <= x1 && iX <= x2)
				bDone = true;
		}
	}
	else
	{
		int iStep;
		int iY;
		bool bDone;
		
		iStep = 1;
		if (y1 > y2)
			iStep = -1;
		
		if (iPenStyle != wxSOLID)
			iStep *= 3;
		
		bDone = false;
		
		for (iY = y1; !bDone; iY += iStep)
		{
			double fX;
			double fWeightLeft;
			double fWeightRight;
			int iLeft;
			int iRight;
			int iShift;
			
			fX = ((double) (iY - y1) / iDY * iDX) + x1;
			
			fWeightRight = fX - floor(fX);
			fWeightLeft = 1 - fWeightRight;
			
			wxASSERT(fWeightRight <= 1);
			wxASSERT(fWeightRight >= 0);

			wxASSERT(fWeightLeft <= 1);
			wxASSERT(fWeightLeft >= 0);
			
			iLeft = -iPenWidth / 2;
			iRight = iLeft + iPenWidth - 1;

			iShift = iLeft;
			ShadePoint((int) floor(fX + iShift), iY, clrPen, fWeightLeft);

			T::SetPen(penCurrent);
			for (iShift++; iShift <= iRight; iShift++)
				T::DrawPoint((int) floor(fX + iShift), iY);
			
			ShadePoint((int) floor(fX + iShift), iY, clrPen, fWeightRight);
			
			if (y2 > y1 && iY >= y2)
				bDone = true;
			if (y2 <= y1 && iY <= y2)
				bDone = true;
		}
	}
	
	T::SetPen(penCurrent);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Replaces default DoDrawPolygon with a bunch of DrawLine calls so polygons
/// are antialiased too.
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::DoDrawPolygon(int n, wxPoint points[],
                               wxCoord xoffset, wxCoord yoffset,
                               int fillStyle)
{
	T::DoDrawPolygon(n, points, xoffset, yoffset, fillStyle);
/*	int i;
	
	if (n == 0)
		return;
	
	wxPen penOld;
	
	penOld = T::GetPen();
	T::SetPen(*wxWHITE_PEN);
	T::DoDrawPolygon(n, points, xoffset, yoffset, fillStyle);
	T::SetPen(penOld);
	
	for (i = 0; i < n - 1; i++)
		T::DrawLine(points[i].x + xoffset, points[i].y + yoffset, points[i + 1].x + xoffset, points[i + 1].y + yoffset);

	T::DrawLine(points[i].x + xoffset, points[i].y + yoffset, points[0].x + xoffset, points[0].y + yoffset);*/
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief Replaces DoDrawText with a call to DoDrawRotatedText
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::DoDrawText(const wxString& text, wxCoord x, wxCoord y)
{
	DoDrawRotatedText(text, x, y, 0);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Draw rotated text.
///
/// (x,y) specify the coordinates of the top left corner of the text 
/// prior to it being rotated. The text is rotated about the point (x,y).
/// angle is specified in degrees where positive degrees rotate the text
/// counterclockwise.
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::DoDrawRotatedText(const wxString& text,
									   wxCoord x, wxCoord y, double angle)
{
	wxMemoryDC dcText;
	wxBitmap bmpText;
	wxCoord cwText;
	wxCoord chText;
	wxFont fntCurrent;	
	wxFont fntBig;
	int iBigWidth;
	int iBigHeight;
	wxBitmap bmpTmp(10, 10);
	
	angle = angle / 180 * PI;
	
	fntCurrent = T::GetFont();
	fntBig.SetFamily(fntCurrent.GetFamily());
	fntBig.SetFaceName(fntCurrent.GetFaceName());
	fntBig.SetStyle(fntCurrent.GetStyle());
	fntBig.SetWeight(fntCurrent.GetWeight());
	fntBig.SetPointSize(fntCurrent.GetPointSize() * m_iAALevel);
	
	dcText.SelectObject(bmpTmp);
	dcText.SetFont(fntBig);
	dcText.GetTextExtent(text, &cwText, &chText);

	iBigWidth = (int) ((fabs(cwText * cos(angle)) + fabs(chText * sin(angle))) * 2);
	iBigHeight = (int) ((fabs(chText * cos(angle)) + fabs(cwText * sin(angle))) * 2);
	
	iBigWidth += m_iAALevel - iBigWidth % m_iAALevel;
	iBigHeight += m_iAALevel - iBigHeight % m_iAALevel;
	
	bmpText.Create(iBigWidth, iBigHeight);
	
	dcText.SelectObject(bmpText);
	
	dcText.SetPen(m_penMask);
	dcText.SetBrush(m_brMask);
	dcText.DrawRectangle(0, 0, iBigWidth, iBigHeight);
	
	dcText.SetTextForeground(T::GetTextForeground());
	dcText.SetTextBackground(T::GetTextBackground());
	
	dcText.DrawRotatedText(text, iBigWidth / 2, iBigHeight / 2, angle * 180 / PI);
	
	dcText.SelectObject(wxNullBitmap);
	
	wxImage imgScaledText = bmpText.ConvertToImage().Rescale(bmpText.GetWidth() / m_iAALevel, bmpText.GetHeight() / m_iAALevel);
	imgScaledText.SetMask(TRUE);
	imgScaledText.SetMaskColour(m_clrMask.Red(), m_clrMask.Green(), m_clrMask.Blue());
	
	T::DrawBitmap(imgScaledText, x - imgScaledText.GetWidth() / 2, y - imgScaledText.GetHeight() / 2, TRUE);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Replaces default DoDrawEllipse with an antialiased version
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::DoDrawEllipse(wxCoord x, wxCoord y, wxCoord width, wxCoord height)
{
	wxMemoryDC dcBig;
	wxBitmap bmpBig;
	wxPen penDrawing;

	bmpBig.Create((width + 1) * m_iAALevel, (height + 1) * m_iAALevel);
	
	dcBig.SelectObject(bmpBig);
	
	dcBig.SetPen(m_penMask);
	dcBig.SetBrush(m_brMask);
	dcBig.DrawRectangle(0, 0, (width + 1) * m_iAALevel, (height + 1) * m_iAALevel);
	
	penDrawing = T::GetPen();
	penDrawing.SetWidth(penDrawing.GetWidth() * m_iAALevel);
	dcBig.SetPen(penDrawing);
	dcBig.SetBrush(T::GetBrush());
	dcBig.DrawEllipse(0, 0, width * m_iAALevel, height * m_iAALevel);
	dcBig.SelectObject(wxNullBitmap);
	
	wxImage imgScaled = bmpBig.ConvertToImage().Rescale(bmpBig.GetWidth() / m_iAALevel, bmpBig.GetHeight() / m_iAALevel);
	imgScaled.SetMask(TRUE);
	imgScaled.SetMaskColour(m_clrMask.Red(), m_clrMask.Green(), m_clrMask.Blue());
	
	T::DrawBitmap(imgScaled, x, y, TRUE);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Computes the text extent for antialiased text (it's not quite the same
/// as non-antialiased)
///
//////////////////////////////////////////////////////////////////////////////
template <class T>
void wxAntiAliasedDC<T>::DoGetTextExtent(const wxString& string,
									 wxCoord *x, wxCoord *y,
									 wxCoord *descent,
									 wxCoord *externalLeading,
									 wxFont *theFont) const
{
	wxFont fntBig;
	wxFont fntCurrent;
	
	fntCurrent = T::GetFont();

	fntBig.SetFamily(fntCurrent.GetFamily());
	fntBig.SetFaceName(fntCurrent.GetFaceName());
	fntBig.SetStyle(fntCurrent.GetStyle());
	fntBig.SetWeight(fntCurrent.GetWeight());
	fntBig.SetPointSize(fntCurrent.GetPointSize() * m_iAALevel);

	wxMemoryDC dcTmp;
	wxBitmap bmpTmp(10, 10);

	dcTmp.SelectObject(bmpTmp);
	dcTmp.SetFont(fntBig);
	dcTmp.GetTextExtent(string, x, y, descent, externalLeading, theFont);

	*x /= m_iAALevel;
	*y /= m_iAALevel;
}

template <class T>

#if wxCHECK_VERSION(2, 8, 0)
void wxAntiAliasedDC<T>::DoSelect(const wxBitmap & bmpObj)
#else
void wxAntiAliasedDC<T>::SelectObject(const wxBitmap & bmpObj)
#endif
{
#if wxCHECK_VERSION(2, 8, 0)
	T::DoSelect(bmpObj);
#else
	T::SelectObject(bmpObj);
#endif

	m_iWidth = bmpObj.GetWidth();
	m_iHeight = bmpObj.GetHeight();

	Setup(m_iWidth, m_iHeight);	
}

template <class T>
wxColour wxAntiAliasedDC<T>::GetPixelColor(wxCoord x, wxCoord y)
{
	int iOffset;

	wxASSERT(x >= 0);
	wxASSERT(y >= 0);
	wxASSERT(x < m_iWidth);
	wxASSERT(y < m_iHeight);
	
	iOffset = m_iWidth * y * 4 + x * 4;
		
	if (!m_pcPixelCache[iOffset])
		return T::GetBackground().GetColour();
		
	return wxColour(m_pcPixelCache[iOffset + 1], m_pcPixelCache[iOffset + 2], m_pcPixelCache[iOffset + 3]);
}

template <class T>
void wxAntiAliasedDC<T>::SetPixelColor(wxCoord x, wxCoord y, const wxColour & c)
{
	int iOffset;
	
	wxASSERT(x >= 0);
	wxASSERT(y >= 0);
	wxASSERT(x < m_iWidth);
	wxASSERT(y < m_iHeight);

	iOffset = m_iWidth * y * 4 + x * 4;
		
	m_pcPixelCache[iOffset] = 1;
	m_pcPixelCache[iOffset + 1] = c.Red();
	m_pcPixelCache[iOffset + 2] = c.Green();
	m_pcPixelCache[iOffset + 3] = c.Blue();
}

template <class T>
void wxAntiAliasedDC<T>::DoDrawLines(int n, wxPoint points[], wxCoord xoffset, wxCoord yoffset)
{
	int i;
	
	if (n == 0)
		return;
	
	for (i = 0; i < n - 1; i++)
		DoDrawLine(points[i].x + xoffset, points[i].y + yoffset, points[i + 1].x + xoffset, points[i + 1].y + yoffset);

	DoDrawLine(points[i].x + xoffset, points[i].y + yoffset, points[0].x + xoffset, points[0].y + yoffset);
}


template <class T>
void wxAntiAliasedDC<T>::Setup(int iWidth, int iHeight)
{
	int i;
	
	m_iWidth = iWidth;
	m_iHeight = iHeight;
	
	if (m_pcPixelCache)
		delete [] m_pcPixelCache;
		
	m_pcPixelCache = new wxUint8[m_iWidth * m_iHeight * 4];
	
	for (i = 0; i < m_iWidth * m_iHeight * 4; i++)
		m_pcPixelCache[i] = 0;
}

#endif

